@@ -16,14 +16,25 @@
  */
 
 #include <sys/utsname.h>
+#include <ctype.h>
+#include <dirent.h>
+#include <errno.h>
+#include <limits.h>
+#include <stdbool.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
+#include <sys/stat.h>
+#include <time.h>
+#include <unistd.h>
 
 #include "queue.h"
 #include "librc.h"
+#include "helpers.h"
+#include "misc.h"
 
 #define GENDEP          RC_LIBEXECDIR "/sh/gendepends.sh"
 
-#define RC_DEPCONFIG    RC_SVCDIR "/depconfig"
-
 static const char *bootlevel = NULL;
 
 static char *
@@ -54,27 +65,23 @@ void
 rc_deptree_free(RC_DEPTREE *deptree)
 {
 	RC_DEPINFO *di;
-	RC_DEPINFO *di2;
+	RC_DEPINFO *di_save;
 	RC_DEPTYPE *dt;
-	RC_DEPTYPE *dt2;
+	RC_DEPTYPE *dt_save;
 
 	if (!deptree)
 		return;
 
-	di = TAILQ_FIRST(deptree);
-	while (di) {
-		di2 = TAILQ_NEXT(di, entries);
-		dt = TAILQ_FIRST(&di->depends);
-		while (dt) {
-			dt2 = TAILQ_NEXT(dt, entries);
+	TAILQ_FOREACH_SAFE(di, deptree, entries, di_save) {
+		TAILQ_FOREACH_SAFE(dt, &di->depends, entries, dt_save) {
+			TAILQ_REMOVE(&di->depends, dt, entries);
 			rc_stringlist_free(dt->services);
 			free(dt->type);
 			free(dt);
-			dt = dt2;
 		}
+		TAILQ_REMOVE(deptree, di, entries);
 		free(di->service);
 		free(di);
-		di = di2;
 	}
 
 	/* Use free() here since rc_deptree_free should not call itself */
@@ -93,6 +100,17 @@ get_depinfo(const RC_DEPTREE *deptree, const char *service)
 	return NULL;
 }
 
+static RC_DEPINFO *
+make_depinfo(RC_DEPTREE *deptree, const char *service)
+{
+	RC_DEPINFO *depinfo = xmalloc(sizeof(*depinfo));
+	TAILQ_INIT(&depinfo->depends);
+	depinfo->service = xstrdup(service);
+	TAILQ_INSERT_TAIL(deptree, depinfo, entries);
+
+	return depinfo;
+}
+
 static RC_DEPTYPE *
 get_deptype(const RC_DEPINFO *depinfo, const char *type)
 {
@@ -106,32 +124,46 @@ get_deptype(const RC_DEPINFO *depinfo, const char *type)
 	return NULL;
 }
 
-RC_DEPTREE *
-rc_deptree_load(void) {
-	return rc_deptree_load_file(RC_DEPTREE_CACHE);
+static RC_DEPTYPE *
+make_deptype(RC_DEPINFO *depinfo, const char *type)
+{
+	RC_DEPTYPE *deptype = xmalloc(sizeof(*deptype));
+	deptype->type = xstrdup(type);
+	deptype->services = rc_stringlist_new();
+	TAILQ_INSERT_TAIL(&depinfo->depends, deptype, entries);
+
+	return deptype;
 }
 
-RC_DEPTREE *
-rc_deptree_load_file(const char *deptree_file)
+#ifdef HAVE_MALLOC_EXTENDED_ATTRIBUTE
+__attribute__ ((malloc (rc_deptree_free, 1)))
+#endif
+static RC_DEPTREE *
+make_deptree(void) {
+	RC_DEPTREE *deptree = xmalloc(sizeof(*deptree));
+	TAILQ_INIT(deptree);
+	return deptree;
+}
+
+static RC_DEPTREE *
+deptree_load_file(int dirfd, const char *pathname)
 {
-	FILE *fp;
 	RC_DEPTREE *deptree;
 	RC_DEPINFO *depinfo = NULL;
 	RC_DEPTYPE *deptype = NULL;
 	char *line = NULL;
-	size_t len = 0;
+	size_t size;
 	char *type;
 	char *p;
 	char *e;
 	int i;
+	FILE *fp;
 
-	if (!(fp = fopen(deptree_file, "r")))
+	if (!(fp = do_fopenat(dirfd, pathname, O_RDONLY)))
 		return NULL;
 
-	deptree = xmalloc(sizeof(*deptree));
-	TAILQ_INIT(deptree);
-	while ((rc_getline(&line, &len, fp)))
-	{
+	deptree = make_deptree();
+	while (xgetline(&line, &size, fp) != -1) {
 		p = line;
 		e = strsep(&p, "_");
 		if (!e || strcmp(e, "depinfo") != 0)
@@ -146,10 +178,7 @@ rc_deptree_load_file(const char *deptree_file)
 			e = get_shell_value(p);
 			if (!e || *e == '\0')
 				continue;
-			depinfo = xmalloc(sizeof(*depinfo));
-			TAILQ_INIT(&depinfo->depends);
-			depinfo->service = xstrdup(e);
-			TAILQ_INSERT_TAIL(deptree, depinfo, entries);
+			depinfo = make_depinfo(deptree, e);
 			deptype = NULL;
 			continue;
 		}
@@ -160,20 +189,28 @@ rc_deptree_load_file(const char *deptree_file)
 		e = get_shell_value(p);
 		if (!e || *e == '\0')
 			continue;
-		if (!deptype || strcmp(deptype->type, type) != 0) {
-			deptype = xmalloc(sizeof(*deptype));
-			deptype->services = rc_stringlist_new();
-			deptype->type = xstrdup(type);
-			TAILQ_INSERT_TAIL(&depinfo->depends, deptype, entries);
-		}
+		if (!deptype || strcmp(deptype->type, type) != 0)
+			deptype = make_deptype(depinfo, type);
 		rc_stringlist_add(deptype->services, e);
 	}
-	fclose(fp);
 	free(line);
+	fclose(fp);
 
 	return deptree;
 }
 
+RC_DEPTREE *
+rc_deptree_load(void)
+{
+	return deptree_load_file(rc_dirfd(RC_DIR_SVCDIR), "deptree");
+}
+
+RC_DEPTREE *
+rc_deptree_load_file(const char *deptree_file)
+{
+	return deptree_load_file(AT_FDCWD, deptree_file);
+}
+
 static bool
 valid_service(const char *runlevel, const char *service, const char *type)
 {
@@ -256,11 +293,11 @@ get_provided1(const char *runlevel, RC_STRINGLIST *providers,
 /* Work out if a service is provided by another service.
    For example metalog provides logger.
    We need to be able to handle syslogd providing logger too.
-   We do this by checking whats running, then what's starting/stopping,
+   We do this by checking what's running, then what's starting/stopping,
    then what's run in the runlevels and finally alphabetical order.
 
    If there are any bugs in rc-depend, they will probably be here as
-   provided dependancy can change depending on runlevel state.
+   provided dependency can change depending on runlevel state.
    */
 static RC_STRINGLIST *
 get_provided(const RC_DEPINFO *depinfo, const char *runlevel, int options)
@@ -547,18 +584,15 @@ rc_deptree_order(const RC_DEPTREE *deptree, const char *runlevel, int options)
    oldest (or newest) found.
 */
 static bool
-deep_mtime_check(const char *target, bool newer,
-	    time_t *rel, char *file)
+deep_mtime_check(int target_dir, const char *target, bool newer, time_t *rel, char *file)
 {
 	struct stat buf;
 	bool retval = true;
 	DIR *dp;
 	struct dirent *d;
-	char path[PATH_MAX];
-	int serrno = errno;
 
 	/* If target does not exist, return true to mimic shell test */
-	if (stat(target, &buf) != 0)
+	if (fstatat(target_dir, target, &buf, 0) != 0)
 		return true;
 
 	if (newer) {
@@ -579,21 +613,18 @@ deep_mtime_check(const char *target, bool newer,
 		}
 	}
 
-	/* If not a dir then reset errno */
-	if (!(dp = opendir(target))) {
-		errno = serrno;
+	if (!S_ISDIR(buf.st_mode) || !(dp = do_opendirat(target_dir, target)))
 		return retval;
-	}
 
 	/* Check all the entries in the dir */
 	while ((d = readdir(dp))) {
 		if (d->d_name[0] == '.')
 			continue;
-		snprintf(path, sizeof(path), "%s/%s", target, d->d_name);
-		if (!deep_mtime_check(path, newer, rel, file)) {
+		if (!deep_mtime_check(dirfd(dp), d->d_name, newer, rel, file)) {
 			retval = false;
 		}
 	}
+
 	closedir(dp);
 	return retval;
 }
@@ -603,8 +634,7 @@ deep_mtime_check(const char *target, bool newer,
  * the return value arguments are non-null).
  */
 static bool
-mtime_check(const char *source, const char *target, bool newer,
-	    time_t *rel, char *file)
+mtime_check(int dirfd, const char *source, const char *target, bool newer, time_t *rel, char *file)
 {
 	struct stat buf;
 	time_t mtime;
@@ -615,7 +645,7 @@ mtime_check(const char *source, const char *target, bool newer,
 		return false;
 	mtime = buf.st_mtime;
 
-	retval = deep_mtime_check(target,newer,&mtime,file);
+	retval = deep_mtime_check(dirfd, target,newer,&mtime,file);
 	if (rel) {
 		*rel = mtime;
 	}
@@ -623,18 +653,16 @@ mtime_check(const char *source, const char *target, bool newer,
 }
 
 bool
-rc_newer_than(const char *source, const char *target,
-	      time_t *newest, char *file)
+rc_newer_than(const char *source, const char *target, time_t *newest, char *file)
 {
 
-	return mtime_check(source, target, true, newest, file);
+	return mtime_check(AT_FDCWD, source, target, true, newest, file);
 }
 
 bool
-rc_older_than(const char *source, const char *target,
-	      time_t *oldest, char *file)
+rc_older_than(const char *source, const char *target, time_t *oldest, char *file)
 {
-	return mtime_check(source, target, false, oldest, file);
+	return mtime_check(AT_FDCWD, source, target, false, oldest, file);
 }
 
 typedef struct deppair
@@ -653,43 +681,20 @@ static const DEPPAIR deppairs[] = {
 	{ NULL, NULL }
 };
 
-static const char *const depdirs[] =
-{
-	RC_SVCDIR,
-	RC_SVCDIR "/starting",
-	RC_SVCDIR "/started",
-	RC_SVCDIR "/stopping",
-	RC_SVCDIR "/inactive",
-	RC_SVCDIR "/wasinactive",
-	RC_SVCDIR "/failed",
-	RC_SVCDIR "/hotplugged",
-	RC_SVCDIR "/daemons",
-	RC_SVCDIR "/options",
-	RC_SVCDIR "/exclusive",
-	RC_SVCDIR "/scheduled",
-	RC_SVCDIR "/tmp",
-	NULL
-};
-
 bool
 rc_deptree_update_needed(time_t *newest, char *file)
 {
 	bool newer = false;
+	const int *dirfds;
 	RC_STRINGLIST *config;
 	RC_STRING *s;
-	int i;
 	struct stat buf;
 	time_t mtime;
 
-	/* Create base directories if needed */
-	for (i = 0; depdirs[i]; i++)
-		if (mkdir(depdirs[i], 0755) != 0 && errno != EEXIST)
-			fprintf(stderr, "mkdir `%s': %s\n", depdirs[i], strerror(errno));
-
 	/* Quick test to see if anything we use has changed and we have
 	 * data in our deptree. */
 
-	if (stat(RC_DEPTREE_CACHE, &buf) == 0) {
+	if (fstatat(rc_dirfd(RC_DIR_SVCDIR), "deptree", &buf, 0) == 0) {
 		mtime = buf.st_mtime;
 	} else {
 		/* No previous cache found.
@@ -700,28 +705,20 @@ rc_deptree_update_needed(time_t *newest, char *file)
 		mtime = time(NULL);
 	}
 
-	newer |= !deep_mtime_check(RC_INITDIR,true,&mtime,file);
-	newer |= !deep_mtime_check(RC_CONFDIR,true,&mtime,file);
-#ifdef RC_PKG_INITDIR
-    newer |= !deep_mtime_check(RC_PKG_INITDIR,true,&mtime,file);
-#endif
-#ifdef RC_PKG_CONFDIR
-    newer |= !deep_mtime_check(RC_PKG_CONFDIR,true,&mtime,file);
-#endif
-#ifdef RC_LOCAL_INITDIRs
-    newer |= !deep_mtime_check(RC_LOCAL_INITDIR,true,&mtime,file);
-#endif
-#ifdef RC_LOCAL_CONFDIR
-    newer |= !deep_mtime_check(RC_LOCAL_CONFDIR,true,&mtime,file);
-#endif
-    newer |= !deep_mtime_check(RC_CONF,true,&mtime,file);
+	for (size_t i = 0, count = rc_scriptdirfds(&dirfds); i < count; i++) {
+		newer |= !deep_mtime_check(dirfds[i], "init.d", true, &mtime, file);
+		newer |= !deep_mtime_check(dirfds[i], "conf.d", true, &mtime, file);
+	}
+
+	newer |= !deep_mtime_check(rc_dirfd(RC_DIR_SYSCONF), "rc.conf", true, &mtime, file);
+	if (rc_is_user())
+		newer |= !deep_mtime_check(rc_dirfd(RC_DIR_USRCONF), "rc.conf", true, &mtime, file);
 
 	/* Some init scripts dependencies change depending on config files
 	 * outside of baselayout, like syslog-ng, so we check those too. */
-	config = rc_config_list(RC_DEPCONFIG);
-	TAILQ_FOREACH(s, config, entries) {
-		newer |= !deep_mtime_check(s->value, true, &mtime, file);
-	}
+	config = config_list(rc_dirfd(RC_DIR_SVCDIR), "depconfig");
+	TAILQ_FOREACH(s, config, entries)
+		newer |= !deep_mtime_check(AT_FDCWD, s->value, true, &mtime, file);
 	rc_stringlist_free(config);
 
 	/* Return newest file time, if requested */
@@ -732,6 +729,33 @@ rc_deptree_update_needed(time_t *newest, char *file)
 	return newer;
 }
 
+static void
+setup_environment(void)
+{
+	struct utsname uts;
+	size_t env_size;
+	char *env;
+	FILE *mem = xopen_memstream(&env, &env_size);
+
+	for (const char * const *dirs = rc_scriptdirs(); *dirs; dirs++) {
+		fputs(*dirs, mem);
+		if (dirs[1])
+			fputc(' ', mem);
+	}
+	xclose_memstream(mem);
+
+	setenv("RC_SCRIPTDIRS", env, 1);
+	free(env);
+
+	/* Some init scripts need RC_LIBEXECDIR to source stuff
+	   Ideally we should be setting our full env instead */
+	if (!getenv("RC_LIBEXECDIR"))
+		setenv("RC_LIBEXECDIR", RC_LIBEXECDIR, 0);
+
+	if (uname(&uts) == 0)
+		setenv("RC_UNAME", uts.sysname, 1);
+}
+
 /* This is a 7 phase operation
    Phase 1 is a shell script which loads each init script and config in turn
    and echos their dependency info to stdout
@@ -746,6 +770,7 @@ rc_deptree_update_needed(time_t *newest, char *file)
 bool
 rc_deptree_update(void)
 {
+
 	FILE *fp;
 	RC_DEPTREE *deptree, *providers;
 	RC_DEPINFO *depinfo = NULL, *depinfo_np, *di;
@@ -753,30 +778,22 @@ rc_deptree_update(void)
 	RC_STRINGLIST *config, *dupes, *types, *sorted, *visited;
 	RC_STRING *s, *s2, *s2_np, *s3, *s4;
 	char *line = NULL;
-	size_t len = 0;
-	char *depend, *depends, *service, *type, *nosys, *onosys;
-	size_t i, k, l;
+	size_t size;
+	char *depend, *depends, *service, *type;
+	size_t i, l;
 	bool retval = true;
 	const char *sys = rc_sys();
-	struct utsname uts;
 	int serrno;
 
-	/* Some init scripts need RC_LIBEXECDIR to source stuff
-	   Ideally we should be setting our full env instead */
-	if (!getenv("RC_LIBEXECDIR"))
-		setenv("RC_LIBEXECDIR", RC_LIBEXECDIR, 0);
-
-	if (uname(&uts) == 0)
-		setenv("RC_UNAME", uts.sysname, 1);
 	/* Phase 1 - source all init scripts and print dependencies */
+	setup_environment();
 	if (!(fp = popen(GENDEP, "r")))
 		return false;
 
-	deptree = xmalloc(sizeof(*deptree));
-	TAILQ_INIT(deptree);
 	config = rc_stringlist_new();
-	while ((rc_getline(&line, &len, fp)))
-	{
+
+	deptree = make_deptree();
+	while (xgetline(&line, &size, fp) != -1) {
 		depends = line;
 		service = strsep(&depends, " ");
 		if (!service || !*service)
@@ -786,12 +803,8 @@ rc_deptree_update(void)
 		if (!depinfo || strcmp(depinfo->service, service) != 0) {
 			deptype = NULL;
 			depinfo = get_depinfo(deptree, service);
-			if (!depinfo) {
-				depinfo = xmalloc(sizeof(*depinfo));
-				TAILQ_INIT(&depinfo->depends);
-				depinfo->service = xstrdup(service);
-				TAILQ_INSERT_TAIL(deptree, depinfo, entries);
-			}
+			if (!depinfo)
+				depinfo = make_depinfo(deptree, service);
 		}
 
 		/* We may not have any depends */
@@ -800,20 +813,16 @@ rc_deptree_update(void)
 
 		/* Get the type */
 		if (strcmp(type, "config") != 0) {
-			if (!deptype || strcmp(deptype->type, type) != 0)
+			if (!deptype || strcmp(deptype->type, type) != 0) {
 				deptype = get_deptype(depinfo, type);
-			if (!deptype) {
-				deptype = xmalloc(sizeof(*deptype));
-				deptype->type = xstrdup(type);
-				deptype->services = rc_stringlist_new();
-				TAILQ_INSERT_TAIL(&depinfo->depends, deptype, entries);
+				if (!deptype)
+					deptype = make_deptype(depinfo, type);
 			}
 		}
 
 		/* Now add each depend to our type.
 		   We do this individually so we handle multiple spaces gracefully */
-		while ((depend = strsep(&depends, " ")))
-		{
+		while ((depend = strsep(&depends, " "))) {
 			if (depend[0] == 0)
 				continue;
 
@@ -822,9 +831,8 @@ rc_deptree_update(void)
 				continue;
 			}
 
-			/* Don't provide ourself */
-			if (strcmp(type, "iprovide") == 0 &&
-			    strcmp(depend, service) == 0)
+			/* Don't depend on ourself */
+			if (strcmp(depend, service) == 0)
 				continue;
 
 			/* .sh files are not init scripts */
@@ -866,7 +874,9 @@ rc_deptree_update(void)
 	/* Phase 2 - if we're a special system, remove services that don't
 	 * work for them. This doesn't stop them from being run directly. */
 	if (sys) {
-		len = strlen(sys);
+		char *nosys, *onosys;
+		size_t len = strlen(sys);
+
 		nosys = xmalloc(len + 2);
 		nosys[0] = '-';
 		for (i = 0; i < len; i++)
@@ -880,29 +890,30 @@ rc_deptree_update(void)
 			onosys[i + 2] = (char)tolower((unsigned char)sys[i]);
 		onosys[i + 2] = '\0';
 
-		TAILQ_FOREACH_SAFE(depinfo, deptree, entries, depinfo_np)
-			if ((deptype = get_deptype(depinfo, "keyword")))
-				TAILQ_FOREACH(s, deptype->services, entries)
-					if (strcmp(s->value, nosys) == 0 ||
-					    strcmp(s->value, onosys) == 0)
-					{
-						provide = get_deptype(depinfo, "iprovide");
-						TAILQ_REMOVE(deptree, depinfo, entries);
-						TAILQ_FOREACH(di, deptree, entries) {
-							TAILQ_FOREACH_SAFE(dt, &di->depends, entries, dt_np) {
-								rc_stringlist_delete(dt->services, depinfo->service);
-								if (provide)
-									TAILQ_FOREACH(s2, provide->services, entries)
-										rc_stringlist_delete(dt->services, s2->value);
-								if (!TAILQ_FIRST(dt->services)) {
-									TAILQ_REMOVE(&di->depends, dt, entries);
-									free(dt->type);
-									free(dt->services);
-									free(dt);
-								}
-							}
+		TAILQ_FOREACH_SAFE(depinfo, deptree, entries, depinfo_np) {
+			if (!(deptype = get_deptype(depinfo, "keyword")))
+				continue;
+			TAILQ_FOREACH(s, deptype->services, entries) {
+				if (strcmp(s->value, nosys) != 0 && strcmp(s->value, onosys) != 0)
+					continue;
+				provide = get_deptype(depinfo, "iprovide");
+				TAILQ_REMOVE(deptree, depinfo, entries);
+				TAILQ_FOREACH(di, deptree, entries) {
+					TAILQ_FOREACH_SAFE(dt, &di->depends, entries, dt_np) {
+						rc_stringlist_delete(dt->services, depinfo->service);
+						if (provide)
+							TAILQ_FOREACH(s2, provide->services, entries)
+								rc_stringlist_delete(dt->services, s2->value);
+						if (!TAILQ_FIRST(dt->services)) {
+							TAILQ_REMOVE(&di->depends, dt, entries);
+							free(dt->type);
+							free(dt->services);
+							free(dt);
 						}
 					}
+				}
+			}
+		}
 		free(nosys);
 		free(onosys);
 	}
@@ -910,24 +921,20 @@ rc_deptree_update(void)
 	/* Phase 3 - add our providers to the tree */
 	providers = xmalloc(sizeof(*providers));
 	TAILQ_INIT(providers);
-	TAILQ_FOREACH(depinfo, deptree, entries)
-		if ((deptype = get_deptype(depinfo, "iprovide")))
-			TAILQ_FOREACH(s, deptype->services, entries) {
-				TAILQ_FOREACH(di, providers, entries)
-					if (strcmp(di->service, s->value) == 0)
-						break;
-				if (!di) {
-					di = xmalloc(sizeof(*di));
-					TAILQ_INIT(&di->depends);
-					di->service = xstrdup(s->value);
-					TAILQ_INSERT_TAIL(providers, di, entries);
-				}
-			}
+	TAILQ_FOREACH(depinfo, deptree, entries) {
+		if (!(deptype = get_deptype(depinfo, "iprovide")))
+			continue;
+		TAILQ_FOREACH(s, deptype->services, entries) {
+			di = get_depinfo(providers, s->value);
+			if (!di)
+				di = make_depinfo(providers, s->value);
+		}
+	}
 	TAILQ_CONCAT(deptree, providers, entries);
 	free(providers);
 
 	/* Phase 4 - backreference our depends */
-	TAILQ_FOREACH(depinfo, deptree, entries)
+	TAILQ_FOREACH(depinfo, deptree, entries) {
 		for (i = 0; deppairs[i].depend; i++) {
 			deptype = get_deptype(depinfo, deppairs[i].depend);
 			if (!deptype)
@@ -936,33 +943,23 @@ rc_deptree_update(void)
 				di = get_depinfo(deptree, s->value);
 				if (!di) {
 					if (strcmp(deptype->type, "ineed") == 0) {
-						fprintf(stderr,
-							 "Service `%s' needs non"
-							 " existent service `%s'\n",
+						fprintf(stderr, "Service '%s' needs non existent service '%s'\n",
 							 depinfo->service, s->value);
 						dt = get_deptype(depinfo, "broken");
-						if (!dt) {
-							dt = xmalloc(sizeof(*dt));
-							dt->type = xstrdup("broken");
-							dt->services = rc_stringlist_new();
-							TAILQ_INSERT_TAIL(&depinfo->depends, dt, entries);
-						}
+						if (!dt)
+							dt = make_deptype(depinfo, "broken");
 						rc_stringlist_addu(dt->services, s->value);
 					}
 					continue;
 				}
 
 				dt = get_deptype(di, deppairs[i].addto);
-				if (!dt) {
-					dt = xmalloc(sizeof(*dt));
-					dt->type = xstrdup(deppairs[i].addto);
-					dt->services = rc_stringlist_new();
-					TAILQ_INSERT_TAIL(&di->depends, dt, entries);
-				}
+				if (!dt)
+					dt = make_deptype(di, deppairs[i].addto);
 				rc_stringlist_addu(dt->services, depinfo->service);
 			}
 		}
-
+	}
 
 	/* Phase 5 - Remove broken before directives */
 	types = rc_stringlist_new();
@@ -1035,45 +1032,40 @@ rc_deptree_update(void)
 	   This works and should be entirely shell parseable provided that depend
 	   names don't have any non shell variable characters in
 	   */
-	if ((fp = fopen(RC_DEPTREE_CACHE, "w"))) {
+	if ((fp = do_fopenat(rc_dirfd(RC_DIR_SVCDIR), "deptree", O_WRONLY | O_CREAT | O_TRUNC))) {
 		i = 0;
 		TAILQ_FOREACH(depinfo, deptree, entries) {
-			fprintf(fp, "depinfo_%zu_service='%s'\n",
-				i, depinfo->service);
+			fprintf(fp, "depinfo_%zu_service='%s'\n", i, depinfo->service);
 			TAILQ_FOREACH(deptype, &depinfo->depends, entries) {
-				k = 0;
+				size_t k = 0;
 				TAILQ_FOREACH(s, deptype->services, entries) {
-					fprintf(fp,
-						"depinfo_%zu_%s_%zu='%s'\n",
-						i, deptype->type, k, s->value);
-					k++;
+					fprintf(fp, "depinfo_%zu_%s_%zu='%s'\n",
+							i, deptype->type, k++, s->value);
 				}
 			}
 			i++;
 		}
 		fclose(fp);
 	} else {
-		fprintf(stderr, "fopen `%s': %s\n",
-			RC_DEPTREE_CACHE, strerror(errno));
+		fprintf(stderr, "fopen '%s/deptree': %s\n", rc_svcdir(), strerror(errno));
 		retval = false;
 	}
 
 	/* Save our external config files to disk */
 	if (TAILQ_FIRST(config)) {
-		if ((fp = fopen(RC_DEPCONFIG, "w"))) {
+		if ((fp = do_fopenat(rc_dirfd(RC_DIR_SVCDIR), "depconfig", O_WRONLY | O_CREAT | O_TRUNC))) {
 			TAILQ_FOREACH(s, config, entries)
 				fprintf(fp, "%s\n", s->value);
 			fclose(fp);
 		} else {
-			fprintf(stderr, "fopen `%s': %s\n",
-				RC_DEPCONFIG, strerror(errno));
+			fprintf(stderr, "fopen '%s/depconfig': %s\n", rc_svcdir(), strerror(errno));
 			retval = false;
 		}
 	} else {
-		unlink(RC_DEPCONFIG);
+		unlinkat(rc_dirfd(RC_DIR_SVCDIR), "depconfig", 0);
 	}
 
 	rc_stringlist_free(config);
-	free(deptree);
+	rc_deptree_free(deptree);
 	return retval;
 }
